
# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
""" Abstract base classes
"""

from pyrates.frontend.yaml import to_dict as dict_from_yaml

__author__ = "Daniel Rose"
__status__ = "Development"


class AbstractBaseTemplate:
    """Abstract base class for templates"""

    target_ir = None  # placeholder for template-specific intermediate representation (IR) target class
    cache = {}  # dictionary that keeps track of already loaded templates

    def __init__(self, name: str, path: str, description: str = "A template."):
        self.name = name
        self.path = path
        self.__doc__ = description  # overwrite class-specific doc with user-defined description

    def __repr__(self):
        return f"<{self.__class__.__name__} '{self.path}'>"

    @staticmethod
    def _complete_template_path(target_path: str, source_path: str) -> str:
        """Check if path contains a folder structure and prepend own path, if it doesn't"""

        if "." not in target_path:
            if "/" in source_path or "\\" in source_path:
                import os
                basedir, _ = os.path.split(source_path)
                target_path = os.path.normpath(os.path.join(basedir, target_path))
            else:
                target_path = ".".join((*source_path.split('.')[:-1], target_path))
        return target_path

    @classmethod
    def from_yaml(cls, path):
        """Convenience method that looks for a loader class for the template type and applies it, assuming
        the class naming convention '<template class>Loader'.

        Parameters:
        -----------
        path
            Path to template in YAML file of form 'directories.file.template'
        """

        if path in cls.cache:
            template = cls.cache[path]
        else:
            template_dict = dict_from_yaml(path)
            try:
                base_path = template_dict.pop("base")
            except KeyError:
                raise KeyError(f"No 'base' defined for template {path}. Please define a "
                               f"base to derive the template from.")
            if base_path == cls.__name__:
                # if base refers to the python representation, instantiate here
                template = cls(**template_dict)
            else:
                # load base if needed
                base_path = cls._complete_template_path(base_path, path)

                base_template = cls.from_yaml(base_path)
                template = base_template.update_template(**template_dict)
                # may fail if "base" is present but empty

            cls.cache[path] = template

        return template

    def update_template(self, *args, **kwargs):
        """Updates the template with a given list of arguments and returns a new instance of the template class."""
        raise NotImplementedError

    def apply(self, *args, **kwargs):
        """Converts the template into its corresponding intermediate representation class."""
        raise NotImplementedError




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network models and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

# system imports
import os
import sys
import json
import time
import argparse
import warnings

# external imports
from numba import njit, config
import numpy as np
import pandas as pd
import scipy.signal as sp
from pyrates.utility.grid_search import grid_search


def cgs_postprocessing(data: np.array):
    """Post processing function that is applied to every model parametrization output.

    Parameters
    ----------
    data
        Simulated model output for a single model parametrization

    Returns
    -------
    np.array
        Post processed model output

    """

    # Add customized post processing here
    # e.g. Computation of power spectral density of the time signal:
    #
    # from scipy.signal import welch
    # dt = data[1] - data[0]
    # f, p  = welch(data.values, fs=1 / dt, axis=0, **kwargs)
    # processed_data = p

    # Placeholder
    processed_data = data

    return processed_data


# Don't make any changes below #
################################

def main(_):
    # tf.config.set_soft_device_placement(True)

    config.THREADING_LAYER = 'omp'

    # Disable general warnings
    warnings.filterwarnings("ignore")

    # disable TF-gpu warnings
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    t_total = time.time()

    # Load command line arguments and create logfile
    ################################################
    print("")
    print("***LOADING COMMAND LINE ARGUMENTS***")
    t0 = time.time()

    config_file = FLAGS.config_file
    subgrid = FLAGS.subgrid
    local_res_file = FLAGS.local_res_file
    build_dir = FLAGS.build_dir

    print(f'Elapsed time: {time.time() - t0:.3f} seconds')

    # Load global config file
    #########################
    print("")
    print("***LOADING GLOBAL CONFIG FILE***")
    t0 = time.time()

    with open(config_file) as g_conf:
        global_config_dict = json.load(g_conf)
        circuit_template = global_config_dict['circuit_template']
        param_map = global_config_dict['param_map']
        dt = global_config_dict['dt']
        simulation_time = global_config_dict['simulation_time']

        # Optional parameters
        #####################
        try:
            sampling_step_size = global_config_dict['sampling_step_size']
        except KeyError:
            sampling_step_size = dt

        try:
            inputs = global_config_dict['inputs']
        except KeyError:
            inputs = {}

        try:
            outputs = global_config_dict['outputs']
        except KeyError:
            outputs = {}

        try:
            init_kwargs = global_config_dict['init_kwargs']
        except KeyError:
            init_kwargs = {}

        print(f'Elapsed time: {time.time() - t0:.3f} seconds')

        # LOAD PARAMETER GRID
        #####################
        print("")
        print("***PREPARING PARAMETER GRID***")
        t0 = time.time()

        # Load subgrid into DataFrame
        param_grid = pd.read_hdf(subgrid, key="subgrid")

        # Drop all columns that don't contain a parameter map value (e.g. status, chunk_idx, err_count) since
        # grid_search() can't handle additional columns
        param_grid = param_grid[list(param_map.keys())]
        print(f'Elapsed time: {time.time() - t0:.3f} seconds')

        # COMPUTE PARAMETER GRID
        ########################
        print("")
        print("***COMPUTING PARAMETER GRID***")
        t0 = time.time()

        results, result_map, t_ = grid_search(
            circuit_template=circuit_template,
            param_grid=param_grid,
            param_map=param_map,
            simulation_time=simulation_time,
            dt=dt,
            sampling_step_size=sampling_step_size,
            permute_grid=False,
            inputs=inputs,
            outputs=outputs.copy(),
            init_kwargs=init_kwargs,
            profile='t',
            build_dir=build_dir,
            njit=True,
            parallel=False)

    print(f'Total parameter grid computation time: {time.time()-t0:.3f} seconds')

    # Post process results and write data to local result file
    ##########################################################
    print("")
    print("***POSTPROCESSING AND CREATING RESULT FILES***")
    t0 = time.time()

    processed_results = pd.DataFrame(data=None, columns=results.columns, index=results.index)
    for idx, circuit in enumerate(result_map.iterrows()):
        circ_idx = result_map.loc[(result_map == tuple(circuit[1].values)).all(1), :].index
        processed_results[circ_idx] = cgs_postprocessing(results[circ_idx].to_numpy())

    with pd.HDFStore(local_res_file, "w") as store:
        store.put(key='results', value=results)
        store.put(key='result_map', value=result_map)

    # TODO: Copy local result file back to master if needed

    print(f'Result files created. Elapsed time: {time.time()-t0:.3f} seconds')
    print("")
    print(f'Total elapsed time: {time.time()-t_total:.3f} seconds')


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--config_file",
        type=str,
        default="/nobackup/spanien1/salomon/CGS/Benchmark_jup/Config/DefaultConfig_5.json",
        help="File to load grid_search configuration parameters from"
    )

    parser.add_argument(
        "--subgrid",
        type=str,
        default="/nobackup/spanien1/salomon/CGS/Benchmark_jup/Grids/Subgrids/DefaultGrid_0/animals/cgs_test_grid.h5",
        help="File to load parameter grid from"
    )

    parser.add_argument(
        "--local_res_file",
        type=str,
        default="/nobackup/spanien1/salomon/WorkerTestData/holgado_subgrid/test_result.h5",
        help="File to save results to"
    )

    parser.add_argument(
        "--build_dir",
        type=str,
        default=os.getcwd(),
        help="Custom PyRates build directory"
    )

    FLAGS = parser.parse_args()

    main(sys.argv)


def postprocessing_1(data_):
    """Compute spike frequency based on frequency in PSD with the highest energy"""

    from pyrates.utility import plot_psd
    import matplotlib.pyplot as plt

    # Store columns for later reconstruction
    cols = data_.columns

    # Plot_psd expects only the out_var as column value
    index_dummy = pd.Index([cols.values[-1][-1]])
    data_.columns = index_dummy
    if not any(np.isnan(data_.values)):
        _ = plot_psd(data_, tmin=1.0, show=False)
        pow_ = plt.gca().get_lines()[-1].get_ydata()
        freqs = plt.gca().get_lines()[-1].get_xdata()
        plt.close()
        max_freq = freqs[np.argmax(pow_)]
        freq_pow = np.max(pow_)
        temp = [max_freq, freq_pow]
        psd = pd.DataFrame(temp, index=['max_freq', 'freq_pow'], columns=cols)
        data_.columns = cols
        return psd

    data_.columns = cols

    # Return empty DataFrame if data contains NaN
    return pd.DataFrame(columns=cols)


def postprocessing_2(data, simulation_time):
    """Compute spike frequency based on average number of spikes (local maxima) per second"""

    import scipy.signal as sp

    cols = data.columns

    np_data = np.array(data.values)

    peaks = sp.argrelextrema(np_data, np.greater)
    num_peaks_temp = int(len(peaks[0]) / simulation_time)

    return pd.DataFrame(num_peaks_temp, index=['num_peaks'], columns=cols)


def postprocessing_3(data, dt):
    """Compute spike frequency based on average time between spikes (local maxima)"""

    import scipy.signal as sp

    cols = data.columns

    np_data = np.array(data.values)

    peaks = sp.argrelextrema(np_data, np.greater)
    diff = np.diff(peaks[0])
    diff = np.mean(diff) * dt
    return 1 / diff




# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""Basic neural mass backend class plus derivations of it.

This module includes the base circuit class that manages the set-up and simulation of population networks. Additionally,
it features various sub-classes that act as circuit constructors and allow to build circuits based on different input
arguments. For more detailed descriptions, see the respective docstrings.

"""

# external packages
from typing import List, Union, Dict, Optional
from copy import deepcopy

# pyrates internal imports
from pyrates import PyRatesException
from pyrates.frontend.template.abc import AbstractBaseTemplate
from pyrates.frontend.template.edge import EdgeTemplate
from pyrates.frontend.template.node import NodeTemplate

# from pyrates.frontend.operator import OperatorTemplate

# meta infos
from pyrates.ir.circuit import CircuitIR
from pyrates.ir.edge import EdgeIR

__author__ = "Richard Gast, Daniel Rose"
__status__ = "Development"


class CircuitTemplate(AbstractBaseTemplate):
    target_ir = CircuitIR

    def __init__(self, name: str, path: str, description: str = "A circuit template.", label: str = "circuit",
                 circuits: dict = None, nodes: dict = None, edges: List[tuple] = None):

        super().__init__(name, path, description)

        self.nodes = {}  # type: Dict[str, NodeTemplate]
        if nodes:
            for key, node in nodes.items():
                if isinstance(node, str):
                    path = self._complete_template_path(node, self.path)
                    self.nodes[key] = NodeTemplate.from_yaml(path)
                else:
                    self.nodes[key] = node

        self.circuits = {}
        if circuits:
            for key, circuit in circuits.items():
                if isinstance(circuit, str):
                    path = self._complete_template_path(circuit, self.path)
                    self.circuits[key] = CircuitTemplate.from_yaml(path)
                else:
                    self.circuits[key] = circuit

        if edges:
            self.edges = self._get_edge_templates(edges)
        else:
            self.edges = []

        self.label = label

    def update_template(self, name: str, path: str, description: str = None,
                        label: str = None, circuits: dict = None, nodes: dict = None,
                        edges: List[tuple] = None):
        """Update all entries of the circuit template in their respective ways."""

        if not description:
            description = self.__doc__

        if not label:
            label = self.label

        if nodes:
            nodes = update_dict(self.nodes, nodes)
        else:
            nodes = self.nodes

        if circuits:
            circuits = update_dict(self.circuits, circuits)
        else:
            circuits = self.circuits

        if edges:
            edges = update_edges(self.edges, edges)
        else:
            edges = self.edges

        return self.__class__(name=name, path=path, description=description,
                              label=label, circuits=circuits, nodes=nodes,
                              edges=edges)

    def apply(self, label: str = None):
        """Create a Circuit graph instance based on the template"""

        if not label:
            label = self.label

        # reformat node templates to NodeIR instances
        nodes = {key: temp.apply() for key, temp in self.nodes.items()}

        # reformat edge templates to EdgeIR instances
        edges = []
        for (source, target, template, values) in self.edges:

            # get edge template and instantiate it
            values = deepcopy(values)
            weight = values.pop("weight", 1.)

            # get delay
            delay = values.pop("delay", None)

            # treat empty dummy edge templates as not existent templates
            if template and len(template.operators) == 0:
                template = None
            if template is None:
                edge_ir = None
                if values:
                    # should not happen. Putting this just in case.
                    raise PyRatesException("An empty edge IR was provided with additional values. "
                                           "No way to figure out where to apply those values.")
            else:
                edge_ir = template.apply(values=values)  # type: Optional[EdgeIR] # edge spec

            edges.append((source, target,  # edge_unique_key,
                          {"edge_ir": edge_ir,
                           "weight": weight,
                           "delay": delay
                           }))

        return CircuitIR(label, self.circuits, nodes, edges, self.path)

    def _get_edge_templates(self, edges: List[Union[tuple, dict]]):
        """
        Reformat edges from [source, target, template_path, variables] to
        [source, target, template_object, variables]

        Parameters
        ----------
        edges

        Returns
        -------
        edges_with_templates
        """
        edges_with_templates = []
        for edge in edges:
            if isinstance(edge, dict):
                edge = (edge["source"], edge["target"], edge["template"], edge["variables"])
            # "template" is EdgeTemplate, just use it
            if isinstance(edge[2], EdgeTemplate):
                temp = edge[2]
            # if not, try to load template path from yaml
            else:
                path = edge[2]
                try:
                    path = self._complete_template_path(path, self.path)
                    temp = EdgeTemplate.from_yaml(path)
                except TypeError:
                    temp = None
            edges_with_templates.append((*edge[0:2], temp, *edge[3:]))
        return edges_with_templates


# def to_circuit(template):
#     return template.apply()


def update_edges(base_edges: List[tuple], updates: List[Union[tuple, dict]]):
    """Add edges to list of edges. Removing or altering is currently not supported."""

    updated = deepcopy(base_edges)
    for edge in updates:
        if isinstance(edge, dict):
            if "variables" in edge:
                edge = [edge["source"], edge["target"], edge["template"], edge["variables"]]
            else:
                edge = [edge["source"], edge["target"], edge["template"]]
        elif not 3 <= len(edge) <= 4:
            raise PyRatesException("Wrong edge data type or not enough arguments")
        updated.append(edge)

    return updated


def update_dict(base_dict: dict, updates: dict):
    updated = deepcopy(base_dict)

    updated.update(updates)

    return updated



# pyrates imports
from pyrates.ir.circuit import CircuitIR
from pyrates.backend import ComputeGraph
from pyrates.utility import plot_connectivity
from pyrates.frontend import CircuitTemplate

# additional imports
import numpy as np
from scipy.spatial.distance import cdist
from scipy.stats.distributions import norm
from pandas import DataFrame
import matplotlib.pyplot as plt
from seaborn import cubehelix_palette
from copy import deepcopy
import time

__author__ = "Richard Gast"
__status__ = "Development"


########################
# parameter definition #
########################

# base parameters
#################

N = 10                  # number of grid points along x/y axis
N2 = int(N*N)
inter_col_dist = 1e-3   # distance between cortical columns in meter
v = 4.0                 # global velocity of axonal signal propagation in meter per second
k = 10.0                # global connection weight scaling

# connectivity parameters
#########################

# initialization
weights = np.zeros((N2, N2))

# create grid to place columns on
grid_points = np.arange(0, N) * inter_col_dist
x, y = np.meshgrid(grid_points, grid_points)
grid_points = []
for i in range(N):
    idx1, idx2 = i, 0
    while idx1 >= 0:
        grid_points.append([x[idx1, idx2], y[idx1, idx2]])
        idx1 -= 1
        idx2 += 1
for i in range(N):
    idx1, idx2 = N-i-1, N-1
    while idx1 <= N-1:
        grid_points.append([x[idx1, idx2], y[idx1, idx2]])
        idx2 -= 1
        idx1 += 1
grid_points = np.asarray(grid_points)
dists = cdist(grid_points, grid_points)
plt.matshow(dists)
plt.show()

# create distance-dependent connectivity
conn_kernel = norm(0., 0.01).pdf(np.linspace)
plt.plot(conn_kernel)
plt.show()




from pyrates.utility import plot_timeseries, grid_search, plot_psd, plot_connectivity
import numpy as np
import matplotlib.pyplot as plt
from seaborn import cubehelix_palette

__author__ = "Richard Gast"
__status__ = "Development"


# parameters
dt = 1e-4
T = 5.
inp = 3. + np.random.randn(int(T/dt), 1) * 1.0

params = {'H_e': np.arange(0.1, 2.01, 0.1)[::-1], 'H_i': np.arange(0.1, 4.01, 0.1)}
param_map = {'H_e': {'var': [('Op_e.0', 'H_e'), ('Op_i.0', 'H_e')],
                     'nodes': ['PC.0', 'IIN.0']},
             'H_i': {'var': [('Op_e.0', 'H_i'), ('Op_i.0', 'H_i')],
                     'nodes': ['PC.0', 'IIN.0']},
             }

# perform simulation
results = grid_search(circuit_template="EI_circuit.Net",
                      param_grid=params, param_map=param_map,
                      inputs={("PC", "Op_e.0", "i_in"): inp}, outputs={"r": ("PC", "Op_e.0", "r")},
                      dt=dt, simulation_time=T, permute_grid=True, sampling_step_size=1e-3)

# plotting
cut_off = 1.
max_freq = np.zeros((len(params['H_e']), len(params['H_i'])))
freq_pow = np.zeros_like(max_freq)
for i, H_e in enumerate(params['H_e']):
    for j, H_i in enumerate(params['H_i']):
        if not results[H_e][H_i].isnull().any().any():
            _ = plot_psd(results[H_e][H_i], tmin=cut_off, show=False)
            pow = plt.gca().get_lines()[-1].get_ydata()
            freqs = plt.gca().get_lines()[-1].get_xdata()
            max_freq[i, j] = freqs[np.argmax(pow)]
            freq_pow[i, j] = np.max(pow)
            plt.close('all')

fig, ax = plt.subplots(ncols=2, figsize=(15, 5), gridspec_kw={})
cm1 = cubehelix_palette(n_colors=int(len(params['H_e'])*len(params['H_i'])), as_cmap=True, start=2.5, rot=-0.1)
cm2 = cubehelix_palette(n_colors=int(len(params['H_e'])*len(params['H_i'])), as_cmap=True, start=-2.0, rot=-0.1)
cax1 = plot_connectivity(max_freq, ax=ax[0], yticklabels=list(np.round(params['H_e'], decimals=2)),
                         xticklabels=list(np.round(params['H_i'], decimals=2)), cmap=cm1)
cax1.set_xlabel('H_i')
cax1.set_ylabel('H_e')
cax1.set_title(f'max freq')
cax2 = plot_connectivity(freq_pow, ax=ax[1], yticklabels=list(np.round(params['H_e'], decimals=2)),
                         xticklabels=list(np.round(params['H_i'], decimals=2)), cmap=cm2)
cax2.set_xlabel('H_i')
cax2.set_ylabel('H_e')
cax2.set_title(f'freq pow')
plt.suptitle('EI-circuit sensitivity to synaptic efficacies (H)')
plt.tight_layout(pad=2.5, rect=(0.01, 0.01, 0.99, 0.96))
fig.savefig("/home/rgast/Documents/Studium/PhD_Leipzig/Figures/BGTCS/eic_efficacies", format="svg")
plt.show()
